//
//      BRAIN 2.3 (Object DataBase with History)
//
// Copyright(c) 2018, Tomas Lachman
// Public domain
//
// *.BDB
//
// 256 bytes reserved ("DNA")
// {
// int T,A,N,L - common header of all objects
//   T - type (0..255 see bdbtypes.txt)
//   A - unused (can be used for access rights or uid of user who modified it)
//   N - next - 0 or pointer to the next version
//   L - last - 0 or pointer to the last version
// char DATA[o_size(T)-16] - content of the object
// }*

#include <stdio.h>
#include <windows.h>

#define MAX_CACHE 1024
#define MAX_SAVE 1024

#define INT4 int
#define UNS4 unsigned int
#define OID int
#define INFINITY 0x7FFFFFFF

#define OT_NULL 0
#define OT_INVALID 1
#define OT_BOOL 2
#define OT_CHAR 3
#define OT_COLOR 4
#define OT_TIME 5
#define OT_DATE 6
#define OT_DATETIME 7
#define OT_INT 8
#define OT_UNS 12
#define OT_SINGLE 16
#define OT_DOUBLE 17
#define OT_DECIMAL 21
#define OT_ACCESSRIGHTS 90
#define OT_TIMESTAMP 91
#define OT_ARRAY 128
#define OT_USER_TYPE 129
#define OT_LIST_BITS 130
#define OT_STRING 131
#define OT_MENU 160
#define OT_COMMAND 161
#define OT_COMBINED 162
#define OT_LIST_SORTED_INT 176
#define OT_LIST_SORTED_OBJ 177


struct TANL_t{ int T,A,N,L; };


FILE *gfo = NULL;
UNS4 giOFSize = 0;
char gchBusy = (char)0;
char gsFirst256[256];

// Exclusive
OID riCurTP = INFINITY;     // CurrentTimePoint - default = INFINITY
OID ri_o_last_oid = 0;      // OID of the last object (get, update)
INT4 ri_o_last_error_code = 0;  // Last error code, 0 == OK
UNS4 ri_o_last_size = 0;        // Size of the last object
INT4 ri_o_last_ver_dif = 0;     // Difference of the returned version to the requested version


// Open file.  IN: asFN  OUT: gfo, giOFSize, gchBusy=false, gsFirst256
int o_init(char asFN[]){

    // Do open file.  IN: asFN  OUT: gfo  MAY: return -1
    if(( gfo = fopen(asFN, "r+b") ) == NULL){
        giOFSize = 0;
        //MessageBox(0, gsBDBNotFound, gsErrorMsg, MB_OK | MB_ICONEXCLAMATION);
        return -1;
    }

    // Get size of the file.  IN: asFN  OUT: giOFSize
    fseek(gfo, 0L, SEEK_END);
    giOFSize = ftell(gfo);
    fseek(gfo, 0L, SEEK_SET);

    // Load DNA
    fread( &gsFirst256, 1, 256, gfo );

    // Init semaphore
    gchBusy = (char)0;

    // return OK
    return 0;
}

// Close file.  IN: gfo  OUT: gfo=NULL
void o_done(){
    if(gfo != NULL){
        fclose(gfo);
        gfo = NULL;
    }
}

// Returns the primary size of type 0..127.
int o_basic_size(int aiTp){
    UNS4 liSize = 0;
    // 32..63: char / unicode
    if( ((aiTp & 127) | 31) == 63){
        return 4;
    }
    // 96..126: byte
    if( ((aiTp & 127) | 31) == 127){
        return 4;
    }
    switch(aiTp & 127){

        case 0:{ liSize=0; break; }    // null
        case 1:{ liSize=0; break; }    // invalid
        case 2:{ liSize=4; break; }    // bool
        case 3:{ liSize=4; break; }    // char
        case 4:{ liSize=4; break; }    // color
        case 5:{ liSize=4; break; }    // time
        case 6:{ liSize=4; break; }    // date
        case 7:{ liSize=8; break; }    // datetime

        // int
        case 8:{ liSize=4; break; }
        case 9:{ liSize=8; break; }
        case 10:{ liSize=16; break; }
        case 11:{ liSize=32; break; }

        // uns
        case 12:{ liSize=4; break; }
        case 13:{ liSize=8; break; }
        case 14:{ liSize=16; break; }
        case 15:{ liSize=32; break; }

        // float
        case 16:{ liSize=4; break; }
        case 17:{ liSize=8; break; }
        case 18:{ liSize=16; break; }
        case 19:{ liSize=32; break; }

        // decimal
        case 20:{ liSize=4; break; }
        case 21:{ liSize=8; break; }
        case 22:{ liSize=16; break; }
        case 23:{ liSize=32; break; }

        // int_vector
        case 24:{ liSize=12; break; }
        case 25:{ liSize=24; break; }
        case 26:{ liSize=48; break; }
        case 27:{ liSize=96; break; }

        // vector
        case 28:{ liSize=12; break; }
        case 29:{ liSize=24; break; }
        case 30:{ liSize=48; break; }
        case 31:{ liSize=96; break; }

        // 32..63: see above

        // pairs
        case 64:{ liSize=8; break; }    // pair for sorted_list with int key
        case 65:{ liSize=8; break; }    // pair for sorted_list with object key
        case 66:{ liSize=8; break; }    // pair for hashed_list with int key
        case 67:{ liSize=8; break; }    // pair for hashed_list with object key

        // special intervals
        case 68:{ liSize=8; break; }
        case 69:{ liSize=8; break; }
        case 70:{ liSize=8; break; }
        case 71:{ liSize=16; break; }

        // angle (rad)
        case 72:{ liSize=4; break; }
        case 73:{ liSize=8; break; }
        case 74:{ liSize=16; break; }
        case 75:{ liSize=32; break; }

        // stangle (srad)
        case 76:{ liSize=8; break; }
        case 77:{ liSize=16; break; }
        case 78:{ liSize=32; break; }
        case 79:{ liSize=64; break; }

        // int_interval
        case 80:{ liSize=8; break; }
        case 81:{ liSize=16; break; }
        case 82:{ liSize=32; break; }
        case 83:{ liSize=64; break; }

        // interval
        case 84:{ liSize=8; break; }
        case 85:{ liSize=16; break; }
        case 86:{ liSize=32; break; }
        case 87:{ liSize=64; break; }

        // int_region
        case 88:{ liSize=24; break; }
        case 89:{ liSize=48; break; }

        // access_rights
        case 90:{ liSize=12; break; }

        // timepoint
        case 91:{ liSize=8; break; }

        // region
        case 92:{ liSize=24; break; }
        case 93:{ liSize=48; break; }
        case 94:{ liSize=96; break; }
        case 95:{ liSize=192; break; }

        // 96..127: see above
    }
    return liSize;
}

// Returns size of the list
int o_list_size(OID type, int elemSize, int count){
    if( (type | 3) == 131 ){
        if(type == 128){ elemSize = elemSize * count + 24; }
        if(type == 129){ elemSize = elemSize * count + 28; }
        if(type == 130){ elemSize = elemSize * ((count + 31) / 32) + 20; }
        if(type == 131){ elemSize = elemSize * ((count + 3) / 4) + 20; }
        //128 list         TANLCB(O)^C
        //129 user_type    TANLCBZ(U)^C
        //130 bits         TANLC(D)^((C+31)/32)
        //131 string       TANLC(D)^((C+3)/4)
    }
    // char array
    else if( (type | 15) == 175 ){
        elemSize = elemSize * ((count + 3) / 4) + 20;
        // 160 = 128+32: menu        TANLC(D)^((C+3)/4)
        // 161 = 128+33: command     TANLC(D)^((C+3)/4)
    }
    // unicode array
    else if( (type | 15) == 191 ){
        elemSize = elemSize * (((count << 1) + 3) / 4) + 20;
    }
    // byte array
    else if( (type | 31) == 255 ){
        elemSize = elemSize * ((count + 3) / 4) + 20;
    }
    // others
    else{
        elemSize = elemSize * count + 20;
    }
    return elemSize;
}

// Returns size of the object
int o_size(DOP pData){
    if(pData == NULL)
        return 0;
    OID liT = *((int*)pData);
    int liSize = 0;
    int liCount = 0;
    if(liT < 256){
        liSize = o_basic_size(liT & 127);
        if(liT < 128){
            liSize += 16;
        }
        else{
            // liT|127==255
            // Lists: size depends on the count of items

            // Get liCount
            liCount = *((int*)(pData+16));

            // Calculate size
            liSize = o_list_size(liT, liSize, liCount);
        }
    }
    else{
        //... :TODO: user types
    }
    return liSize;
}

// Returns object data in a new memory block that must be freed.
// abUseTimePoint: if true, returns a version valid in time riCurTP.
LPSTR o_get(OID aiO, BOOL abUseTimePoint){

    // Variables
    LPSTR lpData = NULL;
    UNS4 liSize;
    INT4 liRead;
    INT4 liCount;
    TANL_t tanl1;

    // File not open (not initialized)
    if(gfo == NULL){ ri_o_last_error_code=101; return NULL; }

    ri_o_last_ver_dif = 0;

    read_o:
    // Can read only objects with OID >= 256
    if(aiO<256){ ri_o_last_error_code=102; return NULL; }

    // OID must be aligned to 32bit
    if((aiO&3)!=0){ ri_o_last_error_code=103; return NULL; }

    // Too big OID
    if(aiO>=giOFSize){ ri_o_last_error_code=104; return NULL; }

    // Load object header.  IN: aiO, gfo  OUT: tanl1
    fpos_t lfp = (fpos_t)aiO;     //>=256 &3=0
    fsetpos(gfo, &lfp);
    liRead = fread(&tanl1, 4, 4, gfo);

    // :TODO: Check access
    if(tanl1.A!=0){ ri_o_last_error_code=105; return NULL; }

    // Find the appropriate time version.
    // IN: abExact, riCurTP, tanl1, aiO, ri_o_last_ver_dif
    // OUT: tanl1, aiO, ri_o_last_ver_dif
    // MAY: goto read_o
    if(abUseTimePoint){
        if(aiO>riCurTP){  // We are in the future

            // L must be less, to prevent infinite loop
            if(tanl1.L>=aiO){ ri_o_last_error_code=106; return NULL; }  // Should never happen

            // Get a previous version (or null if there's no previous version)
            aiO=tanl1.L;
            ri_o_last_ver_dif--;
            goto read_o;
        }
        else{   // We are in the past or presence
            if(tanl1.N!=0){     // 0: Last most version, presence
                if(tanl1.N<=riCurTP){  // Too in the past

                    // N must be greater, to prevent infinite loop
                    if(tanl1.N<=aiO){ ri_o_last_error_code=107; return NULL; }  // Should never happen

                    // Load a next version
                    aiO=tanl1.N;
                    ri_o_last_ver_dif++;
                    goto read_o;
                }
                // aiO <= giCurTP < tanl1.N: Condition of presence
            }
        }
    }

    // Get size of the object.  IN: tanl1, gfo  OUT: liSize  MAY: return NULL
    if(tanl1.T < 256){
        liSize = o_basic_size(tanl1.T & 127);
        if(tanl1.T < 128){
            liSize += 16;
        }
        else{
            // tanl1.T|127==255
            // Lists: size depends on the number of elements

            // Load liCount
            liRead = fread(&liCount, 4, 1, gfo);

            // Calculate size of the whole list
            liSize = o_list_size(tanl1.T, liSize, liCount);
        }
    }
    else{
        // :TODO: types from 256 up
        if(liSize<16){ ri_o_last_error_code=108; return NULL; }
    }

    // Allocate the memory
    // IN: liSize  OUT: lpData  MAY: return NULL
    lpData = LPSTR(malloc(liSize));
    if(lpData == NULL){ ri_o_last_error_code=109; return NULL; }

    // Copy object header to the memory block.  IN: tanl1  OUT: *lpData
    memcpy(lpData, &tanl1, 16);

    if((tanl1.T|127)==255){
        // Copy count
        memcpy( lpData+16, &liCount, 4);

        // Load the rest of data(liSize-20) to lpData[20].  IN: liSize, gfo  OUT: *lpData
        liRead = fread( lpData+20, 1, liSize-20, gfo);
    }
    else{
        // Load the rest of data(liSize-16) do lpData[16].  IN: liSize, gfo  OUT: *lpData
        liRead = fread( lpData+16, 1, liSize-16, gfo);
    }

    ri_o_last_size = liSize;    // Object size
    ri_o_last_oid = aiO;        // Actual OID
    ri_o_last_error_code = 0;   // No error

    //free(lpData);   // Don't forget to free!!!
    return lpData;
}

OID o_update(OID aiOld, LPSTR apData){
    // Check validity of the type(size>0) etc.
    // Find the last time version of the object aiOld(its NEXT is 0),
    // set its NEXT to point to the end of file (this value is returned),
    // set its apData.LAST to the last time version, apData.NEXT=0,
    // append file with apData. The whole operation is exclusive.

    // Semaphore
    if(++gchBusy > 1){ gchBusy--; ri_o_last_error_code=200; return 0; }

    // Variables
    UNS4 liSize;
    INT4 liCount;
    TANL_t tanl1;
    OID liNew = 0;
    fpos_t lfp;

    // ************************* Get object size ***************************

    // File not open (not initialized)
    if(gfo == NULL){ ri_o_last_error_code=201; goto l_return; }

    // Get object size.  IN: apData, gfo  OUT: liSize  MAY: goto l_return
    liSize = o_size(apData);
    if(liSize<16){ ri_o_last_error_code=208; goto l_return; }

    // ********************* Find the last time version ********************

    ri_o_last_ver_dif = 0;

    read_o:
    if(aiOld != 0){
        // OID must be >= 256
        if(aiOld<256){ ri_o_last_error_code=202; goto l_return; }

        // OID must be aligned to 32bit
        if((aiOld&3)!=0){ ri_o_last_error_code=203; goto l_return; }

        // Too big OID
        if(aiOld>=giOFSize){ ri_o_last_error_code=204; goto l_return; }

        // Load object header.  IN: aiOld, gfo  OUT: tanl1  DESTROY: lfp
        lfp = (fpos_t)aiOld;     //>=256 &3=0
        fsetpos(gfo, &lfp);
        fread(&tanl1, 4, 4, gfo);

        // Find the last time version of aiOld object
        if(tanl1.N != 0){

            // Next must be greater (to prevent infinite loop)
            if(tanl1.N <= aiOld){ ri_o_last_error_code=206; goto l_return; }

            // Next time version
            aiOld = tanl1.N;
            ri_o_last_ver_dif++;
            goto read_o;
        }

        // :TODO: check access
        if(tanl1.A != 0){ ri_o_last_error_code=205; goto l_return; }

        // Type of the updated object must be same as that of the updating one
        if(tanl1.T != *((int*)apData)){ ri_o_last_error_code=207; goto l_return; }
    }

    // ************************** Add new object *****************************

    // New object will be saved at the end of file.  IN: giOFSize  OUT: aiNew
    liNew = giOFSize;

    // If there is the last version, set its NEXT pointer to liNew
    // IN: aiOld, aiNew, gfo
    if(aiOld != 0){
        lfp = (fpos_t)(aiOld+8);    // 0 T, 4 A, 8 N, 12 L
        fsetpos(gfo, &lfp);
        fwrite(&liNew, 4, 1, gfo);
    }

    // Set LAST of the new object to aiOld
    memcpy(apData+12, &aiOld, 4);

    // Set NEXT of the new object to null
    *((int*)(apData+8)) = 0;

    // Save apData to file
    // IN: apData, liSize, gfo, aiNew
    lfp = (fpos_t)(liNew);
    fsetpos(gfo, &lfp);
    fwrite(apData, 1, liSize, gfo);

    // Update giOFSize.  IN: liSize, giOFSize  OUT: giOFSize
    giOFSize += liSize;

    ri_o_last_size = liSize;    // Object size
    ri_o_last_oid = liNew;      // Actual OID of the new object
    ri_o_last_error_code = 0;   // No error

    l_return:
    gchBusy--;
    return liNew;
}

// Find object from base and path.
// Default aiBase = 256.
OID o_find(OID aiBase, bool abUseTimePoint, char *apszPath){
    LPSTR lpData = NULL;
    LPSTR lp1 = NULL;
    OID liO = 0;
    int liCount = 0;
    int l1 = 0;
    char lch = 0;

    l_read:
        lch = *apszPath++;
        if(lch != 0){

            // Load data
            lpData = o_get(aiBase, abUseTimePoint);
            if(lpData == NULL){return 0;}

            // Menu
            if(*((int*)lpData) == OT_MENU){

                // MENU: uns4 T,A,N,L,C; char D[C];
                // Where D is like this:
                //   OID mload, INT4 mprm;
                //   (
                //   OID load, INT4 prm;
                //   UNS1 display_text_len; char display_text[display_text_len];
                //   UNS1 command_text_len; char command_text[command_text_len];
                //   )*

                // menu.C
                liCount = *(int*)(lpData+16);

                // pointing to menu.D
                lp1 = lpData + 28;
                liCount -= 28;

                // Find an item with display_text[0] == lch
                // IN: liCount, lp1, lch
                // DESTROY: liCount, lp1
                // MAY: OUT: aiBase, lpData; GOTO: l_read
                while(liCount > 0){
                    if(*(lp1+9) == lch){

                        // Found, set aiBase, tidy up and next char in apszPath
                        aiBase = *((int*)lp1);
                        free(lpData);
                        lpData = NULL;
                        goto l_read;

                    }
                    // Not found, try the next line
                    l1 = (unsigned char)*(lp1+8) + 9;
                    l1 += (unsigned char)*(lp1+l1) + 1;
                    lp1 += liLen;
                    liCount -= liLen;
                }

                // Not found, tidy up and return 0
                free(lpData);
                lpData = NULL;
                return 0;
            }
            // END OT_MENU

            // :TODO: other type?

            // Tidy up
            free(lpData);
            lpData = NULL;
            return 0;
        }
        // END lch != 0

    l_return:
    return aiBase;
}

